clear;
close all;

%%
% choose example 1, 2, or 3
input = 3;

% 0 for explicit gradient descent, 1 for lagged diffusivity,
implicit = 1;

switch input
    case 1
        I = double(imread('cameraman.tif'))/255;
    case 2
        I = double(imread('zebra.png'))/255;
        I = imresize(I, 0.5);
    case 3
        I = double(imread('lion.png'))/255;
        I = imresize(I, 0.5);       
end

% distort with gaussian noise
sigma = 0.1;
I = I + randn(size(I))*sigma;

[ny, nx, nc] = size(I);
N = ny*nx*nc;
f = I(:);

% restoration parameters
lambda_s = 1.0;
lambda_tv = 0.1;
epsilon_gradient = 1e-6;
if implicit
    % large stepsize possible
    dt = 10;
    numit = 20;
    show_every = 1;
else
    % need small stepsize and therefore many iterations
    dt = 0.01;
    numit = 1000;
    show_every = 20;
end

% linear operators
grad = spmat_gradient2d(ny, nx, nc);
div = -grad';

% initialize with distorted image
u_s = f;
u_tv1 = f;
u_tv2 = f;
for i = 1:numit
    
    fprintf("iteration %d\n", i);
 
    % compute gradient of u (flow field inducing the diffusion)
    flow_s   = grad * u_s;
    flow_tv1 = grad * u_tv1;
    flow_tv2 = grad * u_tv2;
    
    % for tv1, normalize the gradient independently for each color channel
    % for tv2, we normalize them jointly (norm of Jacobian)
    flow_tv1 = reshape(flow_tv1, [ny*nx*nc, 2]);
    flow_tv2 = reshape(flow_tv2, [ny*nx, nc*2]);
    
    % compute pointwise norm gradient
    grad_norm_tv1 = sqrt(sum(flow_tv1.^2, 2) + epsilon_gradient);
    grad_norm_tv2 = sqrt(sum(flow_tv2.^2, 2) + epsilon_gradient);

    % do gradient descent step
    if implicit
        % implicit gradient descent (lagged diffusivity)
        
        % diffusivity (non-linear term) fixed based on current u
        g_tv1 = spdiags(repmat(1./grad_norm_tv1, 2, 1), 0, 2*N, 2*N);
        g_tv2 = spdiags(repmat(1./grad_norm_tv2, 6, 1), 0, 2*N, 2*N);
        
        % solve 
        %   ((dt+1)id - dt*lambda*div*g*grad) u_next = u + dt*f
        u_s   = ( (dt+1)*speye(N) - dt*lambda_s *div      *grad ) \ (u_s   + dt*f);
        u_tv1 = ( (dt+1)*speye(N) - dt*lambda_tv*div*g_tv1*grad ) \ (u_tv1 + dt*f);
        u_tv2 = ( (dt+1)*speye(N) - dt*lambda_tv*div*g_tv2*grad ) \ (u_tv2 + dt*f);
        
        % NOTE: For the sobolev reguralizer, the resulting system is linear
        % so you could also solve it exactly in 1 step like done in week 4.
    else   
        % explicit gradient descent
        
        % normalize flow vector field
        flow_tv1 = reshape(flow_tv1 ./ grad_norm_tv1, [2*ny*nx*nc, 1]);
        flow_tv2 = reshape(flow_tv2 ./ grad_norm_tv2, [2*ny*nx*nc, 1]);
        
        u_s   = u_s   - dt * ( (u_s   - f) - lambda_s  * div * flow_s   );
        u_tv1 = u_tv1 - dt * ( (u_tv1 - f) - lambda_tv * div * flow_tv1 );
        u_tv2 = u_tv2 - dt * ( (u_tv2 - f) - lambda_tv * div * flow_tv2 );
    end
    
    % reshape as image and display
    R_s   = reshape(u_s,   [ny,nx,nc]);
    R_tv1 = reshape(u_tv1, [ny,nx,nc]);
    R_tv2 = reshape(u_tv2, [ny,nx,nc]);

    if rem(i, show_every) == 0 || i == numit
        subplot(2,2,1);
        imshow(I);
        title("Noisy image");

        subplot(2,2,2);
        imshow(R_s);
        title(sprintf("Sobolev / Dirichlet (it %d)", i));

        subplot(2,2,3);
        imshow(R_tv1);
        title(sprintf("TV independent color (it %d)", i));
        
        subplot(2,2,4);
        imshow(R_tv2);
        title(sprintf("TV coupled color (it %d)", i));

        drawnow;
    end
end
